from typing import Callable
from typing import Dict
from typing import List

from cleo.commands.command import Command

from ..exceptions import CommandNotFoundException
from .command_loader import CommandLoader


class FactoryCommandLoader(CommandLoader):
    """
    A simple command loader using factories to instantiate commands lazily.
    """

    def __init__(self, factories: Dict[str, Callable]) -> None:
        self._factories = factories

    @property
    def names(self) -> List[str]:
        return list(self._factories.keys())

    def has(self, name: str) -> bool:
        return name in self._factories

    def get(self, name: str) -> Command:
        if name not in self._factories:
            raise CommandNotFoundException(name)

        factory = self._factories[name]

        return factory()



from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from cleo.application import Application
from cleo.commands.command import Command
from cleo.exceptions import CommandNotFoundException


class ApplicationDescription(object):

    GLOBAL_NAMESPACE = "_global"

    def __init__(
        self,
        application: Application,
        namespace: Optional[str] = None,
        show_hidden: bool = False,
    ) -> None:
        self._application: Application = application
        self._namespace = namespace
        self._show_hidden = show_hidden
        self._namespaces: Dict[str, Dict[str, Union[str, List[Command]]]] = {}
        self._commands = {}
        self._aliases = {}

        self._inspect_application()

    @property
    def namespaces(self) -> Dict[str, Dict[str, Union[str, List[Command]]]]:
        return self._namespaces

    @property
    def commands(self) -> Dict[str, Command]:
        return self._commands

    def command(self, name: str) -> Command:
        if name not in self._commands and name not in self._aliases:
            raise CommandNotFoundException(name)

        return self._commands.get(name, self._aliases.get(name))

    def _inspect_application(self):
        namespace = None
        if self._namespace:
            namespace = self._application.find_namespace(self._namespace)

        all_commands = self._application.all(namespace)

        for namespace, commands in self._sort_commands(all_commands):
            names = []

            for name, command in commands:
                if not command.name or command.hidden:
                    continue

                if command.name == name:
                    self._commands[name] = command
                else:
                    self._aliases[name] = command

                names.append(name)

            self._namespaces[namespace] = {"id": namespace, "commands": names}

    def _sort_commands(
        self, commands: Dict[str, Command]
    ) -> List[Tuple[str, List[Tuple[str, Command]]]]:
        """
        Sorts command in alphabetical order
        """
        namespaced_commands = {}
        for name, command in commands.items():
            key = self._application.extract_namespace(name, 1)
            if not key:
                key = "_global"

            if key in namespaced_commands:
                namespaced_commands[key][name] = command
            else:
                namespaced_commands[key] = {name: command}

        for namespace, commands in namespaced_commands.items():
            namespaced_commands[namespace] = sorted(
                commands.items(), key=lambda x: x[0]
            )

        namespaced_commands = sorted(namespaced_commands.items(), key=lambda x: x[0])

        return namespaced_commands



import itertools
import math
import re

from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

from cleo.formatters.formatter import Formatter
from cleo.io.io import IO
from cleo.io.outputs.output import Output

from .table_cell import TableCell
from .table_cell_style import TableCellStyle
from .table_separator import TableSeparator
from .table_style import TableStyle


_Row = List[Union[str, TableCell]]
_Rows = List[Union[_Row, TableSeparator]]


class Table:

    SEPARATOR_TOP: int = 0
    SEPARATOR_TOP_BOTTOM: int = 1
    SEPARATOR_MID: int = 2
    SEPARATOR_BOTTOM: int = 3

    BORDER_OUTSIDE: int = 0
    BORDER_INSIDE: int = 1

    _styles: Optional[Dict[str, TableStyle]] = None

    def __init__(self, io: Union[IO, Output], style: str = None) -> None:
        self._io = io

        if style is None:
            style = "default"

        self._header_title = None
        self._footer_title = None

        self._headers = []

        self._rows = []
        self._horizontal = False

        self._effective_column_widths: Dict[int, int] = {}

        self._number_of_columns = None

        self._column_styles: Dict[int, TableStyle] = {}
        self._column_widths: Dict[int, int] = {}
        self._column_max_widths: Dict[int, int] = {}

        self._rendered = False

        self._style: Optional[TableStyle] = None
        self._init_styles()
        self.set_style(style)

    @property
    def style(self) -> TableStyle:
        return self._style

    def set_style(self, name: str) -> "Table":
        self._init_styles()

        self._style = self._resolve_style(name)

        return self

    def column_style(self, column_index: int) -> TableStyle:
        if column_index in self._column_styles:
            return self._column_styles[column_index]

        return self._style

    def set_column_style(
        self, column_index: int, style: Union[str, TableStyle]
    ) -> "Table":
        self._column_styles[column_index] = self._resolve_style(style)

        return self

    def set_column_width(self, column_index: int, width: int) -> "Table":
        self._column_widths[column_index] = width

        return self

    def set_column_widths(self, widths: List[int]) -> "Table":
        self._column_widths = {}

        for i, width in enumerate(widths):
            self._column_widths[i] = width

        return self

    def set_column_max_width(self, column_index: int, width: int) -> "Table":
        self._column_widths[column_index] = width

        return self

    def set_headers(self, headers: List[str]) -> "Table":
        if headers and not isinstance(headers[0], list):
            headers = [headers]

        self._headers = headers

        return self

    def set_rows(self, rows: _Rows) -> "Table":
        self._rows = []

        return self.add_rows(rows)

    def add_rows(self, rows: _Rows) -> "Table":
        for row in rows:
            self.add_row(row)

        return self

    def add_row(self, row: Union[_Row, TableSeparator]) -> "Table":
        if isinstance(row, TableSeparator):
            self._rows.append(row)

            return self

        self._rows.append(row)

        return self

    def set_header_title(self, header_title: str) -> "Table":
        self._header_title = header_title

        return self

    def set_footer_title(self, footer_title: str) -> "Table":
        self._footer_title = footer_title

        return self

    def horizontal(self, horizontal: bool = True) -> "Table":
        self._horizontal = horizontal

        return self

    def render(self) -> None:
        divider = TableSeparator()

        if self._horizontal:
            rows = []
            headers = self._headers[0] if self._headers else []
            for i, header in enumerate(headers):
                rows.append([header])
                for row in self._rows:
                    if isinstance(row, TableSeparator):
                        continue

                    if len(row) > i:
                        rows[i].append(row[i])
                    elif isinstance(rows[i][0], TableCell) and rows[i][0].colspan >= 2:
                        # There is a title
                        pass
                    else:
                        rows[i].append(None)
        else:
            rows = self._headers + [divider] + self._rows

        self._calculate_number_of_columns(rows)
        rows = list(self._build_table_rows(rows))
        self._calculate_column_widths(rows)

        is_header = not self._horizontal
        is_first_row = self._horizontal

        for row in rows:
            if row is divider:
                is_header = False
                is_first_row = True

                continue

            if isinstance(row, TableSeparator):
                self._render_row_separator()

                continue

            if not row:
                continue

            if is_header or is_first_row:
                if is_first_row:
                    self._render_row_separator(self.SEPARATOR_TOP_BOTTOM)
                    is_first_row = False
                else:
                    self._render_row_separator(
                        self.SEPARATOR_TOP,
                        self._header_title,
                        self._style.header_title_format,
                    )

            if self._horizontal:
                self._render_row(
                    row, self._style.cell_row_format, self._style.cell_header_format
                )
            else:
                self._render_row(
                    row,
                    self._style.cell_header_format
                    if is_header
                    else self._style.cell_row_format,
                )

        self._render_row_separator(
            self.SEPARATOR_BOTTOM,
            self._footer_title,
            self._style.footer_title_format,
        )

        self._cleanup()
        self._rendered = True

    def _render_row_separator(
        self,
        type: int = SEPARATOR_MID,
        title: Optional[str] = None,
        title_format: Optional[str] = None,
    ) -> None:
        """
        Renders horizontal header separator.

        Example:

            +-----+-----------+-------+
        """
        count = self._number_of_columns
        if not count:
            return

        borders = self._style.border_chars
        if not borders[0] and not borders[2] and not self._style.crossing_char:
            return

        crossings = self._style.crossing_chars
        if type == self.SEPARATOR_MID:
            horizontal, left_char, mid_char, right_char = (
                borders[2],
                crossings[8],
                crossings[0],
                crossings[4],
            )
        elif type == self.SEPARATOR_TOP:
            horizontal, left_char, mid_char, right_char = (
                borders[0],
                crossings[1],
                crossings[2],
                crossings[3],
            )
        elif type == self.SEPARATOR_TOP_BOTTOM:
            horizontal, left_char, mid_char, right_char = (
                borders[0],
                crossings[9],
                crossings[10],
                crossings[11],
            )
        else:
            horizontal, left_char, mid_char, right_char = (
                borders[0],
                crossings[7],
                crossings[6],
                crossings[5],
            )

        markup = left_char
        for column in range(count):
            markup += horizontal * self._effective_column_widths[column]
            markup += right_char if column == count - 1 else mid_char

        if title is not None:
            formatted_title = title_format.format(title)
            title_length = len(self._io.remove_format(formatted_title))
            markup_length = len(markup)
            limit = markup_length - 4

            if title_length > limit:
                title_length = limit
                format_length = len(self._io.remove_format(title_format.format("")))
                formatted_title = title_format.format(
                    title[: limit - format_length - 3] + "..."
                )

            title_start = (markup_length - title_length) // 2
            markup = (
                markup[:title_start]
                + formatted_title
                + markup[title_start + title_length :]
            )

        self._io.write_line(self._style.border_format.format(markup))

    def _render_column_separator(self, type: int = BORDER_OUTSIDE) -> str:
        """
        Renders vertical column separator.
        """
        borders = self._style.border_chars

        return self._style.border_format.format(
            borders[1] if type == self.BORDER_OUTSIDE else borders[3]
        )

    def _render_row(
        self, row: List[str], cell_format: str, first_cell_format: Optional[str] = None
    ) -> None:
        """
        Renders table row.

        Example:

            | 9971-5-0210-0 | A Tale of Two Cities  | Charles Dickens  |
        """
        row_content = self._render_column_separator(self.BORDER_OUTSIDE)
        columns = self._get_row_columns(row)
        last = len(columns) - 1
        for i, column in enumerate(columns):
            if first_cell_format and i == 0:
                row_content += self._render_cell(row, column, first_cell_format)
            else:
                row_content += self._render_cell(row, column, cell_format)

            row_content += self._render_column_separator(
                self.BORDER_OUTSIDE if i == last else self.BORDER_INSIDE
            )

        self._io.write_line(row_content)

    def _render_cell(self, row: _Row, column: int, cell_format: str) -> str:
        """
        Renders a table cell with padding.
        """
        try:
            cell = row[column]
        except IndexError:
            cell = ""

        width = self._effective_column_widths[column]
        if isinstance(cell, TableCell) and cell.colspan > 1:
            # add the width of the following columns(numbers of colspan).
            for next_column in range(column + 1, column + cell.colspan):
                width += (
                    self._get_column_separator_width()
                    + self._effective_column_widths[next_column]
                )

        style = self.column_style(column)

        if isinstance(cell, TableSeparator):
            return style.border_format.format(style.border_chars[2] * width)

        width += len(cell) - len(self._io.remove_format(cell))
        content = style.cell_row_content_format.format(cell)

        pad = style.pad
        if isinstance(cell, TableCell) and isinstance(cell.style, TableCellStyle):
            is_not_styled_by_tag = not re.match(
                r"^<(\w+|(\w+=[\w,]+;?)*)>.+</(\w+|(\w+=\w+;?)*)?>$", str(cell)
            )
            if is_not_styled_by_tag:
                cell_format = cell.style.cell_format
                if cell_format is None:
                    cell_format = f"<{cell.style.tag}>{{}}</>"

                if "</>" in content:
                    content = content.replace("</>", "")
                    width -= 3

                if "<fg=default;bg=default>" in content:
                    content = content.replace("<fg=default;bg=default>")
                    width -= len("<fg=default;bg=default>")

            pad = cell.style.pad

        return cell_format.format(pad(content, width, style.padding_char))

    def _calculate_number_of_columns(self, rows: _Rows) -> None:
        columns = [0]
        for row in rows:
            if isinstance(row, TableSeparator):
                continue

            columns.append(self._get_number_of_columns(row))

        self._number_of_columns = max(columns)

    def _build_table_rows(self, rows: _Rows) -> Generator:
        unmerged_rows = {}
        row_key = 0
        while row_key < len(rows):
            rows = self._fill_next_rows(rows, row_key)

            # Remove any new line breaks and replace it with a new line
            for column, cell in enumerate(rows[row_key]):
                colspan = cell.colspan if isinstance(cell, TableCell) else 1

                if column in self._column_max_widths and self._column_max_widths[
                    column
                ] < len(self._io.remove_format(cell)):
                    cell = self._io.formatter.format_and_wrap(
                        cell, self._column_max_widths[column] * colspan
                    )

                if "\n" not in cell:
                    continue

                escaped = "\n".join(
                    Formatter.escape_trailing_backslash(c) for c in cell.split("\n")
                )
                cell = (
                    TableCell(escaped, colspan=cell.colspan)
                    if isinstance(cell, TableCell)
                    else escaped
                )
                lines = cell.replace("\n", "<fg=default;bg=default>\n</>").split("\n")

                for line_key, line in enumerate(lines):
                    if colspan > 1:
                        line = TableCell(line, colspan=colspan)

                    if line_key == 0:
                        rows[row_key][column] = line
                    else:
                        if row_key not in unmerged_rows:
                            unmerged_rows[row_key] = {}

                        if line_key not in unmerged_rows[row_key]:
                            unmerged_rows[row_key][line_key] = self._copy_row(
                                rows, row_key
                            )

                        unmerged_rows[row_key][line_key][column] = line

            row_key += 1

        for row_key, row in enumerate(rows):
            yield self._fill_cells(row)

            if row_key in unmerged_rows:
                for unmerged_row in unmerged_rows[row_key].values():
                    yield self._fill_cells(unmerged_row)

    def _calculate_row_count(self) -> int:
        number_of_rows = len(
            list(
                self._build_table_rows(self._headers + [TableSeparator()] + self._rows)
            )
        )

        if self._headers:
            number_of_rows += 1

        if len(self._rows) > 0:
            number_of_rows += 1

        return number_of_rows

    def _fill_next_rows(self, rows: _Rows, line: int) -> _Rows:
        """
        Fill rows that contains rowspan > 1.
        """
        unmerged_rows = {}

        for column, cell in enumerate(rows[line]):
            if isinstance(cell, TableCell) and cell.rowspan > 1:
                nb_lines = cell.rowspan - 1
                lines = [cell]
                if "\n" in cell:
                    lines = cell.replace("\n", "<fg=default;bg=default>\n</>").split(
                        "\n"
                    )
                    if len(lines) > nb_lines:
                        nb_lines = cell.count("\n")

                    rows[line][column] = TableCell(
                        lines[0], colspan=cell.colspan, style=cell.style
                    )

                # Create a two dimensional dict (rowspan x colspan)
                placeholder = dict(
                    [(k, {}) for k in range(line + 1, line + 1 + nb_lines)]
                )
                for k, v in unmerged_rows.items():
                    if k in placeholder:
                        for l, m in unmerged_rows[k].items():
                            if l in placeholder[k]:
                                placeholder[k][l].update(m)
                            else:
                                placeholder[k][l] = m
                    else:
                        placeholder[k] = v

                unmerged_rows = placeholder

                for unmerged_row_key, _ in unmerged_rows.items():
                    value = ""
                    if unmerged_row_key - line < len(lines):
                        value = lines[unmerged_row_key - line]

                    unmerged_rows[unmerged_row_key][column] = TableCell(
                        value, colspan=cell.colspan, style=cell.style
                    )
                    if nb_lines == unmerged_row_key - line:
                        break

        for unmerged_row_key, unmerged_row in unmerged_rows.items():
            # we need to know if unmerged_row will be merged or inserted into rows
            if (
                unmerged_row_key < len(rows)
                and isinstance(rows[unmerged_row_key], list)
                and (
                    (
                        self._get_number_of_columns(rows[unmerged_row_key])
                        + self._get_number_of_columns(
                            list(unmerged_rows[unmerged_row_key].values())
                        )
                    )
                    <= self._number_of_columns
                )
            ):
                # insert cell into row at cell_key position
                for cell_key, cell in unmerged_row.items():
                    rows[unmerged_row_key].insert(cell_key, cell)
            else:
                row = self._copy_row(rows, unmerged_row_key - 1)
                for column, cell in unmerged_row.items():
                    if len(cell):
                        row[column] = unmerged_row[column]

                rows.insert(unmerged_row_key, row)

        return rows

    def _fill_cells(self, row: _Row) -> List[Union[str, TableCell]]:
        """
        Fills cells for a row that contains colspan > 1.
        """
        new_row = []

        for column, cell in enumerate(row):
            new_row.append(cell)

            if isinstance(cell, TableCell) and cell.colspan > 1:
                for _ in range(column + 1, column + cell.colspan):
                    # insert empty value at column position
                    new_row.append("")

        if new_row:
            return new_row

        return row

    def _copy_row(self, rows: _Rows, line: int) -> _Row:
        """
        Copies a row.
        """
        row = [x for x in rows[line]]

        for cell_key, cell_value in enumerate(row):
            row[cell_key] = ""
            if isinstance(cell_value, TableCell):
                row[cell_key] = TableCell("", colspan=cell_value.colspan)

        return row

    def _get_number_of_columns(self, row: _Row):
        """
        Gets number of columns by row.
        """
        columns = len(row)
        for column in row:
            if isinstance(column, TableCell):
                columns += column.colspan - 1

        return columns

    def _get_row_columns(self, row: _Row) -> List[int]:
        """
        Gets list of columns for the given row.
        """
        columns = list(range(0, self._number_of_columns))

        for cell_key, cell in enumerate(row):
            if isinstance(cell, TableCell) and cell.colspan > 1:
                # exclude grouped columns.
                columns = [
                    x
                    for x in columns
                    if x not in list(range(cell_key + 1, cell_key + cell.colspan))
                ]

        return columns

    def _calculate_column_widths(self, rows: _Rows) -> None:
        """
        Calculates column widths.
        """
        for column in range(0, self._number_of_columns):
            lengths = [0]
            for row in rows:
                if isinstance(row, TableSeparator):
                    continue

                row_ = row.copy()
                for i, cell in enumerate(row_):
                    if isinstance(cell, TableCell):
                        text_content = self._io.remove_format(cell)
                        text_length = len(text_content)
                        if text_length:
                            length = math.ceil(text_length / cell.colspan)
                            content_columns = [
                                text_content[i : i + length]
                                for i in range(0, text_length, length)
                            ]

                            for position, content in enumerate(content_columns):
                                try:
                                    row_[i + position] = content
                                except IndexError:
                                    row_.append(content)

                lengths.append(self._get_cell_width(row_, column))

            self._effective_column_widths[column] = (
                max(lengths) + len(self._style.cell_row_content_format) - 2
            )

    def _get_column_separator_width(self) -> int:
        return len(self._style.border_format.format(self._style.border_chars[3]))

    def _get_cell_width(self, row: _Row, column: int) -> int:
        """
        Gets cell width.
        """
        cell_width = 0

        try:
            cell = row[column]
            cell_width = len(self._io.remove_format(cell))
        except IndexError:
            pass

        column_width = (
            self._column_widths[column] if column in self._column_widths else 0
        )
        cell_width = max(cell_width, column_width)

        if column in self._column_max_widths:
            return min(self._column_max_widths[column], cell_width)

        return cell_width

    def _cleanup(self):
        self._column_widths = {}
        self._number_of_columns = None

    @classmethod
    def _init_styles(cls) -> None:
        if cls._styles is not None:
            return

        borderless = TableStyle()
        borderless.set_horizontal_border_chars("=")
        borderless.set_vertical_border_chars(" ")
        borderless.set_default_crossing_char(" ")

        compact = TableStyle()
        compact.set_horizontal_border_chars("")
        compact.set_vertical_border_chars(" ")
        compact.set_default_crossing_char("")
        compact.set_cell_row_content_format("{}")

        box = TableStyle()
        box.set_horizontal_border_chars("─")
        box.set_vertical_border_chars("│")
        box.set_crossing_chars("┼", "┌", "┬", "┐", "┤", "┘", "┴", "└", "├")

        box_double = TableStyle()
        box_double.set_horizontal_border_chars("═", "─")
        box_double.set_vertical_border_chars("║", "│")
        box_double.set_crossing_chars(
            "┼", "╔", "╤", "╗", "╢", "╝", "╧", "╚", "╟", "╠", "╪", "╣"
        )

        cls._styles = {
            "default": TableStyle(),
            "borderless": borderless,
            "compact": compact,
            "box": box,
            "box-double": box_double,
        }

    @classmethod
    def _resolve_style(cls, name: Union[str, TableStyle]) -> TableStyle:
        if isinstance(name, TableStyle):
            return name

        if name in cls._styles:
            return cls._styles[name]

        raise ValueError(f'Table style "{name}" is not defined.')



import ast
import inspect
import io
import keyword
import os
import re
import sys
import tokenize

from typing import List
from typing import Optional
from typing import Union

from _pytest._code.code import Frame
from crashtest.frame_collection import FrameCollection
from crashtest.inspector import Inspector
from crashtest.solution_providers.solution_provider_repository import (
    SolutionProviderRepository,
)

from cleo.formatters.formatter import Formatter
from cleo.io.io import IO
from cleo.io.outputs.output import Output


class Highlighter:

    TOKEN_DEFAULT = "token_default"
    TOKEN_COMMENT = "token_comment"
    TOKEN_STRING = "token_string"
    TOKEN_NUMBER = "token_number"
    TOKEN_KEYWORD = "token_keyword"
    TOKEN_BUILTIN = "token_builtin"
    TOKEN_OP = "token_op"
    LINE_MARKER = "line_marker"
    LINE_NUMBER = "line_number"

    DEFAULT_THEME = {
        TOKEN_STRING: "fg=yellow;options=bold",
        TOKEN_NUMBER: "fg=blue;options=bold",
        TOKEN_COMMENT: "fg=default;options=dark,italic",
        TOKEN_KEYWORD: "fg=magenta;options=bold",
        TOKEN_BUILTIN: "fg=default;options=bold",
        TOKEN_DEFAULT: "fg=default",
        TOKEN_OP: "fg=default;options=dark",
        LINE_MARKER: "fg=red;options=bold",
        LINE_NUMBER: "fg=default;options=dark",
    }

    KEYWORDS = set(keyword.kwlist)
    BUILTINS = set(
        __builtins__.keys() if type(__builtins__) is dict else dir(__builtins__)
    )

    UI = {
        False: {"arrow": ">", "delimiter": "|"},
        True: {"arrow": "→", "delimiter": "│"},
    }

    def __init__(self, supports_utf8: bool = True) -> None:
        self._theme = self.DEFAULT_THEME.copy()
        self._ui = self.UI[supports_utf8]

    def code_snippet(
        self, source: str, line: int, lines_before: int = 2, lines_after: int = 2
    ) -> List[str]:
        token_lines = self.highlighted_lines(source)
        token_lines = self.line_numbers(token_lines, line)

        offset = line - lines_before - 1
        offset = max(offset, 0)
        length = lines_after + lines_before + 1
        token_lines = token_lines[offset : offset + length]

        return token_lines

    def highlighted_lines(self, source):
        source = source.replace("\r\n", "\n").replace("\r", "\n")

        return self.split_to_lines(source)

    def split_to_lines(self, source):
        lines = []
        current_line = 1
        current_col = 0
        buffer = ""
        current_type = None
        source_io = io.BytesIO(source.encode())
        formatter = Formatter()

        def readline():
            return formatter.format(
                formatter.escape(source_io.readline().decode())
            ).encode()

        tokens = tokenize.tokenize(readline)
        line = ""
        for token_info in tokens:
            token_type, token_string, start, end, _ = token_info
            lineno = start[0]
            if lineno == 0:
                # Encoding line
                continue

            if token_type == tokenize.ENDMARKER:
                # End of source
                line += "<{}>{}</>".format(self._theme[current_type], buffer)
                lines.append(line)
                break

            if lineno > current_line:
                diff = lineno - current_line
                if diff > 1:
                    lines += [""] * (diff - 1)

                line += "<{}>{}</>".format(
                    self._theme[current_type], buffer.rstrip("\n")
                )

                # New line
                lines.append(line)
                line = ""
                current_line = lineno
                current_col = 0
                buffer = ""

            if token_string in self.KEYWORDS:
                new_type = self.TOKEN_KEYWORD
            elif token_string in self.BUILTINS or token_string == "self":
                new_type = self.TOKEN_BUILTIN
            elif token_type == tokenize.STRING:
                new_type = self.TOKEN_STRING
            elif token_type == tokenize.NUMBER:
                new_type = self.TOKEN_NUMBER
            elif token_type == tokenize.COMMENT:
                new_type = self.TOKEN_COMMENT
            elif token_type == tokenize.OP:
                new_type = self.TOKEN_OP
            elif token_type == tokenize.NEWLINE:
                continue
            else:
                new_type = self.TOKEN_DEFAULT

            if current_type is None:
                current_type = new_type

            if start[1] > current_col:
                buffer += token_info.line[current_col : start[1]]

            if current_type != new_type:
                line += "<{}>{}</>".format(self._theme[current_type], buffer)
                buffer = ""
                current_type = new_type

            if lineno < end[0]:
                # The token spans multiple lines
                lines.append(line)
                token_lines = token_string.split("\n")
                for token_line in token_lines[1:-1]:
                    lines.append(
                        "<{}>{}</>".format(self._theme[current_type], token_line)
                    )

                current_line = end[0]
                buffer = token_lines[-1][: end[1]]
                line = ""
                continue

            buffer += token_string
            current_col = end[1]
            current_line = lineno

        return lines

    def line_numbers(
        self, lines: List[str], mark_line: Optional[int] = None
    ) -> List[str]:
        max_line_length = max(3, len(str(len(lines))))

        snippet_lines = []
        marker = "<{}>{}</> ".format(self._theme[self.LINE_MARKER], self._ui["arrow"])
        no_marker = "  "
        for i, line in enumerate(lines):
            if mark_line is not None:
                if mark_line == i + 1:
                    snippet = marker
                else:
                    snippet = no_marker

            line_number = "{:>{}}".format(i + 1, max_line_length)
            snippet += "<{}>{}</><{}>{}</> {}".format(
                "fg=default;options=bold"
                if mark_line == i + 1
                else self._theme[self.LINE_NUMBER],
                line_number,
                self._theme[self.LINE_NUMBER],
                self._ui["delimiter"],
                line,
            )
            snippet_lines.append(snippet)

        return snippet_lines


class ExceptionTrace:
    """
    Renders the trace of an exception.
    """

    THEME = {
        "comment": "<fg=black;options=bold>",
        "keyword": "<fg=yellow>",
        "builtin": "<fg=blue>",
        "literal": "<fg=magenta>",
    }

    AST_ELEMENTS = {
        "builtins": __builtins__.keys()
        if type(__builtins__) is dict
        else dir(__builtins__),
        "keywords": [
            getattr(ast, cls)
            for cls in dir(ast)
            if keyword.iskeyword(cls.lower())
            and inspect.isclass(getattr(ast, cls))
            and issubclass(getattr(ast, cls), ast.AST)
        ],
    }

    _FRAME_SNIPPET_CACHE = {}

    def __init__(
        self,
        exception: Exception,
        solution_provider_repository: Optional[SolutionProviderRepository] = None,
    ) -> None:
        self._exception = exception
        self._solution_provider_repository = solution_provider_repository
        self._exc_info = sys.exc_info()
        self._ignore = None

    def ignore_files_in(self, ignore: str) -> "ExceptionTrace":
        self._ignore = ignore

        return self

    def render(self, io: Union[IO, Output], simple: bool = False) -> None:
        if simple:
            io.write_line("")
            io.write_line("<error>{}</error>".format(str(self._exception)))
            return

        return self._render_exception(io, self._exception)

    def _render_exception(self, io: Union[IO, Output], exception: Exception) -> None:
        from crashtest.inspector import Inspector

        inspector = Inspector(exception)
        if not inspector.frames:
            return

        if inspector.has_previous_exception():
            self._render_exception(io, inspector.previous_exception)
            io.write_line("")
            io.write_line(
                "The following error occurred when trying to handle this error:"
            )
            io.write_line("")

        self._render_trace(io, inspector.frames)

        self._render_line(
            io, "<error>{}</error>".format(inspector.exception_name), True
        )
        io.write_line("")
        exception_message = (
            Formatter().format(inspector.exception_message).replace("\n", "\n  ")
        )
        self._render_line(io, "<b>{}</b>".format(exception_message))

        current_frame = inspector.frames[-1]
        self._render_snippet(io, current_frame)

        self._render_solution(io, inspector)

    def _render_snippet(self, io: Union[IO, Output], frame: Frame):
        self._render_line(
            io,
            "at <fg=green>{}</>:<b>{}</b> in <fg=cyan>{}</>".format(
                self._get_relative_file_path(frame.filename),
                frame.lineno,
                frame.function,
            ),
            True,
        )

        code_lines = Highlighter(supports_utf8=io.supports_utf8()).code_snippet(
            frame.file_content, frame.lineno, 4, 4
        )

        for code_line in code_lines:
            self._render_line(io, code_line, indent=4)

    def _render_solution(self, io: Union[IO, Output], inspector: Inspector):
        if self._solution_provider_repository is None:
            return

        solutions = self._solution_provider_repository.get_solutions_for_exception(
            inspector.exception
        )
        symbol = "•"
        if not io.output.supports_utf8():
            symbol = "*"

        for solution in solutions:
            title = solution.solution_title
            description = solution.solution_description
            links = solution.documentation_links

            description = description.replace("\n", "\n    ").strip(" ")

            self._render_line(
                io,
                "<fg=blue;options=bold>{} </><fg=default;options=bold>{}</>: {}{}".format(
                    symbol,
                    title.rstrip("."),
                    description,
                    ",".join("\n    <fg=blue>{}</>".format(link) for link in links),
                ),
                True,
            )

    def _render_trace(self, io: Union[IO, Output], frames: FrameCollection) -> None:
        stack_frames = FrameCollection()
        for frame in frames:
            if (
                self._ignore
                and re.match(self._ignore, frame.filename)
                and not io.is_debug()
            ):
                continue

            stack_frames.append(frame)

        remaining_frames_length = len(stack_frames) - 1
        if io.is_verbose() and remaining_frames_length:
            self._render_line(io, "<fg=yellow>Stack trace</>:", True)
            max_frame_length = len(str(remaining_frames_length))
            frame_collections = stack_frames.compact()
            i = remaining_frames_length
            for collection in frame_collections:
                if collection.is_repeated():
                    if len(collection) > 1:
                        frames_message = "<fg=yellow>{}</> frames".format(
                            len(collection)
                        )
                    else:
                        frames_message = "frame"

                    self._render_line(
                        io,
                        "<fg=blue>{:>{}}</>  Previous {} repeated <fg=blue>{}</> times".format(
                            "...",
                            max_frame_length,
                            frames_message,
                            collection.repetitions + 1,
                        ),
                        True,
                    )

                    i -= len(collection) * (collection.repetitions + 1)

                for frame in collection:
                    relative_file_path = self._get_relative_file_path(frame.filename)
                    relative_file_path_parts = relative_file_path.split(os.path.sep)
                    relative_file_path = "{}".format(
                        "<fg=default;options=dark>{}</>".format(
                            Formatter.escape(os.sep)
                        ).join(
                            relative_file_path_parts[:-1]
                            + [
                                "<fg=default;options=bold>{}</>".format(
                                    relative_file_path_parts[-1]
                                )
                            ]
                        ),
                    )

                    self._render_line(
                        io,
                        "<fg=yellow>{:>{}}</>  {}<fg=default;options=dark>:</><b>{}</b> in <fg=cyan>{}</>".format(
                            i,
                            max_frame_length,
                            relative_file_path,
                            frame.lineno,
                            frame.function,
                        ),
                        True,
                    )

                    if io.is_debug():
                        if (frame, 2, 2) not in self._FRAME_SNIPPET_CACHE:
                            code_lines = Highlighter(
                                supports_utf8=io.supports_utf8()
                            ).code_snippet(
                                frame.file_content,
                                frame.lineno,
                            )

                            self._FRAME_SNIPPET_CACHE[(frame, 2, 2)] = code_lines

                        code_lines = self._FRAME_SNIPPET_CACHE[(frame, 2, 2)]

                        for code_line in code_lines:
                            self._render_line(
                                io,
                                "{:>{}}{}".format(" ", max_frame_length, code_line),
                                indent=3,
                            )
                    else:
                        highlighter = Highlighter(supports_utf8=io.supports_utf8())
                        try:
                            code_line = highlighter.highlighted_lines(
                                frame.line.strip()
                            )[0]
                        except tokenize.TokenError:
                            code_line = frame.line.strip()

                        self._render_line(
                            io,
                            "{:>{}}    {}".format(
                                " ",
                                max_frame_length,
                                code_line,
                            ),
                        )

                    i -= 1

    def _render_line(
        self, io: Union[IO, Output], line: str, new_line: bool = False, indent: int = 2
    ) -> None:
        if new_line:
            io.write_line("")

        io.write_line("{}{}".format(indent * " ", line))

    def _get_relative_file_path(self, filepath: str) -> str:
        cwd = os.getcwd()

        if cwd:
            filepath = filepath.replace(cwd + os.path.sep, "")

        home = os.path.expanduser("~")
        if home:
            filepath = filepath.replace(home + os.path.sep, "~" + os.path.sep)

        return filepath



from typing import Any
from typing import Optional

from cleo.application import Application
from cleo.commands.command import Command
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.definition import Definition
from cleo.io.inputs.option import Option
from cleo.io.io import IO
from cleo.io.outputs.output import Type


class Descriptor:
    def __init__(self) -> None:
        self._io: Optional[IO] = None

    def describe(self, io: IO, obj: Any, **options: Any) -> None:
        self._io = io

        if isinstance(obj, Argument):
            self._describe_argument(obj, **options)
        elif isinstance(obj, Option):
            self._describe_option(obj, **options)
        elif isinstance(obj, Definition):
            self._describe_definition(obj, **options)
        elif isinstance(obj, Command):
            self._describe_command(obj, **options)
        elif isinstance(obj, Application):
            self._describe_application(obj, **options)

    def _write(self, content: str, decorated: bool = True) -> None:
        self._io.write(
            content, new_line=False, type=Type.NORMAL if decorated else Type.RAW
        )

    def _describe_argument(self, argument: Argument, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_option(self, option: Option, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_definition(self, definition: Definition, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_command(self, command: Command, **options: Any) -> None:
        raise NotImplementedError()

    def _describe_application(self, application: Application, **options: Any) -> None:
        raise NotImplementedError()



import json
import re

from typing import Any
from typing import List
from typing import Union

from cleo.application import Application
from cleo.commands.command import Command
from cleo.formatters.formatter import Formatter
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.definition import Definition
from cleo.io.inputs.option import Option

from .descriptor import Descriptor


class TextDescriptor(Descriptor):
    def _describe_argument(self, argument: Argument, **options: Any) -> None:
        if argument.default is not None and (
            not isinstance(argument.default, list) or argument.default
        ):
            default = "<comment> [default: {}]</comment>".format(
                self._format_default_value(argument.default)
            )
        else:
            default = ""

        total_width = options.get("total_width", len(argument.name))

        spacing_width = total_width - len(argument.name)
        self._write(
            "  <c1>{}</c1>  {}{}{}".format(
                argument.name,
                " " * spacing_width,
                re.sub(
                    r"\s*[\r\n]\s*",
                    "\n" + " " * (total_width + 4),
                    argument.description,
                ),
                default,
            )
        )

    def _describe_option(self, option: Option, **options: Any) -> None:
        if (
            option.accepts_value()
            and option.default is not None
            and (not isinstance(option.default, list) or option.default)
        ):
            default = "<comment> [default: {}]</comment>".format(
                self._format_default_value(option.default)
            )
        else:
            default = ""

        value = ""
        if option.accepts_value():
            value = "=" + option.name.upper()

            if not option.requires_value():
                value = "[" + value + "]"

        total_width = options.get(
            "total_width", self._calculate_total_width_for_options([option])
        )

        synopsis = "{}{}".format(
            "-{}, ".format(option.shortcut) if option.shortcut else "    ",
            "--{}{}".format(option.name, value),
        )

        spacing_width = total_width - len(synopsis)

        self._write(
            "  <c1>{}</c1>  {}{}{}{}".format(
                synopsis,
                " " * spacing_width,
                re.sub(
                    r"\s*[\r\n]\s*",
                    "\n" + " " * (total_width + 4),
                    option.description,
                ),
                default,
                "<comment> (multiple values allowed)</comment>"
                if option.is_list()
                else "",
            )
        )

    def _describe_definition(self, definition: Definition, **options: Any) -> None:
        arguments = definition.arguments
        definition_options = definition.options
        total_width = self._calculate_total_width_for_options(definition_options)

        for argument in arguments:
            total_width = max(total_width, len(argument.name))

        if arguments:
            self._write("<b>Arguments:</b>")
            self._write("\n")

            for argument in arguments:
                self._describe_argument(argument, total_width=total_width)
                self._write("\n")

        if arguments and definition_options:
            self._write("\n")

        if definition_options:
            later_options = []

            self._write("<b>Options:</b>")

            for option in definition_options:
                if option.shortcut and len(option.shortcut) > 1:
                    later_options.append(option)
                    continue

                self._write("\n")
                self._describe_option(option, total_width=total_width)

            for option in later_options:
                self._write("\n")
                self._describe_option(option, total_width=total_width)

    def _describe_command(self, command: Command, **options: Any) -> None:
        command.merge_application_definition(False)

        description = command.description
        if description:
            self._write("<b>Description:</b>")
            self._write("\n")
            self._write("  " + description)
            self._write("\n\n")

        self._write("<b>Usage:</b>")
        for usage in [command.synopsis(True)] + command.aliases + command.usages:
            self._write("\n")
            self._write("  " + Formatter.escape(usage))

        self._write("\n")

        definition = command.definition
        if definition.options or definition.arguments:
            self._write("\n")
            self._describe_definition(definition, **options)
            self._write("\n")

        help_text = command.processed_help
        if help_text and help_text != description:
            self._write("\n")
            self._write("<b>Help:</b>")
            self._write("\n")
            self._write("  " + help_text.replace("\n", "\n  "))
            self._write("\n")

    def _describe_application(self, application: Application, **options: Any) -> None:
        from .application_description import ApplicationDescription

        described_namespace = options.get("namespace")
        description = ApplicationDescription(application, namespace=described_namespace)

        help_text = application.help
        if help_text:
            self._write("{}\n\n".format(help_text))

        self._write("<b>Usage:</b>\n")
        self._write("  command [options] [arguments]\n\n")

        self._describe_definition(Definition(application.definition.options), **options)

        self._write("\n")
        self._write("\n")

        commands = description.commands
        namespaces = description.namespaces

        if described_namespace and namespaces:
            described_namespace_info = list(namespaces.values())[0]
            for name in described_namespace_info["commands"]:
                commands[name] = description.command(name)

        # calculate max width based on available commands per namespace
        all_commands = list(commands.keys())
        for namespace in namespaces.values():
            all_commands += namespace["commands"]

        width = self._get_column_width(all_commands)
        if described_namespace:
            self._write(
                '<b>Available commands for the "{}" namespace:</b>'.format(
                    described_namespace
                )
            )
        else:
            self._write("<b>Available commands:</b>")

        for namespace in namespaces.values():
            namespace["commands"] = [c for c in namespace["commands"] if c in commands]

            if not namespace["commands"]:
                continue

            if (
                not described_namespace
                and namespace["id"] != ApplicationDescription.GLOBAL_NAMESPACE
            ):
                self._write("\n")
                self._write(" <comment>{}</comment>".format(namespace["id"]))

            for name in namespace["commands"]:
                self._write("\n")
                spacing_width = width - len(name)
                command = commands[name]
                command_aliases = (
                    self._get_command_aliases_text(command)
                    if command.name == name
                    else ""
                )
                self._write(
                    "  <c1>{}</c1>{}{}".format(
                        name, " " * spacing_width, command_aliases + command.description
                    )
                )

            self._write("\n")

    def _format_default_value(self, default: Any) -> str:
        if isinstance(default, str):
            default = Formatter.escape(default)
        elif isinstance(default, list):
            new_default = []
            for value in default:
                if isinstance(value, str):
                    new_default.append(Formatter.escape(value))

            default = new_default
        elif isinstance(default, dict):
            new_default = {}
            for key, value in default.items():
                if isinstance(value, str):
                    new_default[key] = Formatter.escape(value)

            default = new_default

        return json.dumps(default).replace("\\\\", "\\")

    def _calculate_total_width_for_options(self, options: List[Option]) -> int:
        total_width = 0

        for option in options:
            name_length = 1 + max(len(option.shortcut or ""), 1) + 4 + len(option.name)

            if option.accepts_value():
                value_length = 1 + len(option.name)
                if not option.requires_value():
                    value_length += 2

                name_length += value_length

            total_width = max(total_width, name_length)

        return total_width

    def _get_column_width(self, commands: List[Union[Command, str]]) -> int:
        widths = []

        for command in commands:
            if isinstance(command, Command):
                widths.append(len(command.name))
                for alias in command.aliases:
                    widths.append(len(alias))
            else:
                widths.append(len(command))

        if not widths:
            return 0

        return max(widths) + 2

    def _get_command_aliases_text(self, command: Command) -> str:
        text = ""
        aliases = command.aliases

        if aliases:
            text = "[" + "|".join(aliases) + "] "

        return text



import os
import re

from collections import namedtuple
from typing import Any
from typing import Dict
from typing import List

from cleo.io.inputs.argument import Argument
from cleo.io.inputs.option import Option


class Parser(object):
    @classmethod
    def parse(cls, expression: str) -> Dict[str, Any]:
        """
        Parse the given console command definition into a dict.
        """
        parsed = {"name": None, "arguments": [], "options": []}

        if not expression.strip():
            raise ValueError("Console command signature is empty.")

        expression = expression.replace(os.linesep, "")

        matches = re.match(r"[^\s]+", expression)

        if not matches:
            raise ValueError("Unable to determine command name from signature.")

        name = matches.group(0)
        parsed["name"] = name

        tokens = re.findall(r"\{\s*(.*?)\s*\}", expression)

        if tokens:
            parsed.update(cls._parameters(tokens))

        return parsed

    @classmethod
    def _parameters(cls, tokens: List[str]) -> Dict[str, Any]:
        """
        Extract all of the parameters from the tokens.
        """
        arguments = []
        options = []

        for token in tokens:
            if not token.startswith("--"):
                arguments.append(cls._parse_argument(token))
            else:
                options.append(cls._parse_option(token))

        return {"arguments": arguments, "options": options}

    @classmethod
    def _parse_argument(cls, token: str) -> Argument:
        """
        Parse an argument expression.
        """
        description = ""
        validator = None

        if " : " in token:
            token, description = tuple(token.split(" : ", 2))

            token = token.strip()

            description = description.strip()

        # Checking validator:
        matches = re.match(r"(.*)\((.*?)\)", token)
        if matches:
            token = matches.group(1).strip()
            validator = matches.group(2).strip()

        if token.endswith("?*"):
            return Argument(
                token.rstrip("?*"),
                required=False,
                is_list=True,
                description=description,
            )
        elif token.endswith("*"):
            return Argument(
                token.rstrip("*"),
                is_list=True,
                description=description,
            )
        elif token.endswith("?"):
            return Argument(
                token.rstrip("?"),
                required=False,
                description=description,
            )

        matches = re.match(r"(.+)=(.+)", token)
        if matches:
            return Argument(
                matches.group(1),
                required=False,
                description=description,
                default=matches.group(2),
            )

        return Argument(
            token,
            description=description,
        )

    @classmethod
    def _parse_option(cls, token: str) -> Option:
        """
        Parse an option expression.
        """
        description = ""
        validator = None

        if " : " in token:
            token, description = tuple(token.split(" : ", 2))

            token = token.strip()

            description = description.strip()

        # Checking validator:
        matches = re.match(r"(.*)\((.*?)\)", token)
        if matches:
            token = matches.group(1).strip()
            validator = matches.group(2).strip()

        shortcut = None

        matches = re.split(r"\s*\|\s*", token, 2)

        if len(matches) > 1:
            shortcut = matches[0].lstrip("-")
            token = matches[1]
        else:
            token = token.lstrip("-")

        default = None
        flag = True
        requires_value = False
        is_list = False

        if token.endswith("=*"):
            flag = False
            is_list = True
            requires_value = True
            token = token.rstrip("=*")
        elif token.endswith("=?*"):
            flag = False
            is_list = True
            token = token.rstrip("=?*")
        elif token.endswith("=?"):
            flag = False
            token = token.rstrip("=?")
        elif token.endswith("="):
            flag = False
            requires_value = True
            token = token.rstrip("=")
        else:
            matches = re.match(r"(.+)(=[?*]*)(.+)", token)
            if matches:
                flag = False
                token = matches.group(1)
                operator = matches.group(2)
                default = matches.group(3)

                if operator == "=*":
                    requires_value = True
                    is_list = True
                elif operator == "=?*":
                    is_list = True
                elif operator == "=?":
                    requires_value = False
                elif operator == "=":
                    requires_value = True

        return Option(
            token,
            shortcut,
            flag=flag,
            requires_value=requires_value,
            is_list=is_list,
            description=description,
            default=default,
        )



from typing import Optional

from .inputs.input import Input
from .inputs.string_input import StringInput
from .io import IO
from .outputs.null_output import NullOutput


class NullIO(IO):
    def __init__(self, input: Optional[Input] = None) -> None:
        if input is None:
            input = StringInput("")

        super().__init__(input, NullOutput(), NullOutput())



import re

from typing import Any
from typing import List
from typing import Optional
from typing import Union

from cleo.exceptions import ValueException
from cleo.io.io import IO

from .question import Question


class SelectChoiceValidator:
    def __init__(self, question: Question) -> None:
        """
        Constructor.
        """
        self._question = question
        self._values = question.choices

    def validate(self, selected: Union[str, int]) -> str:
        """
        Validate a choice.
        """
        # Collapse all spaces.
        if isinstance(selected, int):
            selected = str(selected)

        selected_choices = selected.replace(" ", "")

        if self._question.supports_multiple_choices():
            # Check for a separated comma values
            if not re.match("^[a-zA-Z0-9_-]+(?:,[a-zA-Z0-9_-]+)*$", selected_choices):
                raise ValueException(self._question.error_message.format(selected))

            selected_choices = selected_choices.split(",")
        else:
            selected_choices = [selected]

        multiselect_choices = []
        for value in selected_choices:
            results = []

            for key, choice in enumerate(self._values):
                if choice == value:
                    results.append(key)

            if len(results) > 1:
                raise ValueException(
                    "The provided answer is ambiguous. Value should be one of {}.".format(
                        " or ".join(str(r) for r in results)
                    )
                )

            try:
                result = self._values.index(value)
                result = self._values[result]
            except ValueError:
                try:
                    value = int(value)

                    if 0 <= value < len(self._values):
                        result = self._values[value]
                    else:
                        result = False
                except ValueError:
                    result = False

            if result is False:
                raise ValueException(self._question.error_message.format(value))

            multiselect_choices.append(result)

        if self._question.supports_multiple_choices():
            return multiselect_choices

        return multiselect_choices[0]


class ChoiceQuestion(Question):
    """
    Multiple choice question.
    """

    def __init__(
        self, question: str, choices: List[str], default: Optional[Any] = None
    ) -> None:
        super(ChoiceQuestion, self).__init__(question, default)

        self._multi_select = False
        self._choices = choices
        self._validator = SelectChoiceValidator(self).validate
        self._autocomplete_values = choices
        self._prompt = " > "
        self._error_message = 'Value "{}" is invalid'

    @property
    def error_message(self) -> str:
        return self._error_message

    @property
    def choices(self) -> List[str]:
        return self._choices

    def supports_multiple_choices(self) -> bool:
        return self._multi_select

    def set_multi_select(self, multi_select: bool) -> None:
        self._multi_select = multi_select

    def set_error_message(self, message: str) -> None:
        self._error_message = message

    def _write_prompt(self, io: IO) -> None:
        """
        Outputs the question prompt.
        """
        message = self._question
        default = self._default

        if default is None:
            message = "<question>{}</question>: ".format(message)
        elif self._multi_select:
            choices = self._choices
            default = default.split(",")

            for i, value in enumerate(default):
                default[i] = choices[int(value.strip())]

            message = "<question>{}</question> [<comment>{}</comment>]:".format(
                message, ", ".join(default)
            )
        else:
            choices = self._choices
            message = "<question>{}</question> [<comment>{}</comment>]:".format(
                message, choices[int(default)]
            )

        if len(self._choices) > 1:
            width = max(*map(len, [str(k) for k, _ in enumerate(self._choices)]))
        else:
            width = 1

        messages = [message]
        for key, value in enumerate(self._choices):
            messages.append(" [<comment>{:{}}</>] {}".format(key, width, value))

        io.write_error_line("\n".join(messages))

        message = self._prompt

        io.write_error(message)
